/*
 * Copyright (C) 2002 Roman Zippel <zippel@linux-m68k.org>
 * Released under the terms of the GNU GPL v2.0.
 */

#include <stdlib.h>
#include <string.h>
#include "zconf.tab.h"
#include "zconf.h"

struct menu rootmenu;
struct menu *current_menu, *current_entry;
static struct menu **last_entry_ptr;
static struct statement *first_stmt;

struct file *file_list;
struct file *current_file;

void menu_init(void)
{
	current_menu = &rootmenu;
	last_entry_ptr = &rootmenu.list;
	current_entry = NULL;
}

void menu_add_entry(struct symbol *sym)
{
	struct menu *menu;

	menu = malloc(sizeof(*menu));
	memset(menu, 0, sizeof(*menu));
	menu->sym = sym;
	menu->parent = current_menu;
	menu->file = current_file;
	menu->lineno = zconf_lineno();

	*last_entry_ptr = menu;
	last_entry_ptr = &menu->next;
	current_entry = menu;

	if (current_menu->sym && sym_is_choice(current_menu->sym))
		/* choice menu block */
		menu->dep = expr_alloc_symbol(current_menu->sym);
		//menu->dep = expr_copy(current_menu->dep);
	else
		/* normal menu block */
		menu->dep = expr_copy(E_EXPR(current_menu->prompt->visible));
}

void menu_end_entry(void)
{
	struct statement *stmt;

	/* add dep to all stmt of this menu entry */
	for (stmt = first_stmt; stmt; stmt = stmt->next)
		E_EXPR(stmt->visible) = expr_alloc_and(expr_copy(current_entry->dep), E_EXPR(stmt->visible));

	/* clean up */
	first_stmt = NULL;
}

void menu_add_menu(void)
{
	current_menu = current_entry;
	last_entry_ptr = &current_entry->list;
}

void menu_end_menu(void)
{
	last_entry_ptr = &current_menu->next;
	current_menu = current_menu->parent;
}

void menu_add_dep(struct expr *dep)
{
	current_entry->dep = expr_alloc_and(current_entry->dep, dep);
}

void menu_set_type(int type)
{
	struct symbol *sym = current_entry->sym;

	if (sym->type == type)
		return;
	if (sym->type == SYMBOL_UNKNOWN) {
		sym->type = type;
		return;
	}
	fprintf(stderr, "%s:%d: type of '%s' redefined from '%s' to '%s'\n",
		current_entry->file->name, current_entry->lineno,
		sym->name ? sym->name : "<choice>", sym_type_name(sym->type), sym_type_name(type));
}

struct statement *create_stmt(int token)
{
	struct statement *stmt;

	stmt = malloc(sizeof(*stmt));
	memset(stmt, 0, sizeof(*stmt));
	stmt->token = token;
	stmt->file = current_file;
	stmt->lineno = zconf_lineno();
	if (current_file) {
		if (!current_file->stmt)
			current_file->stmt = stmt;
		if (current_file->last_stmt)
			current_file->last_stmt->next_pos = stmt;
		current_file->last_stmt = stmt;
	}

	return stmt;
}

struct statement *menu_add_stmt(int token, char *prompt, struct symbol *def, struct expr *dep)
{
	struct statement *stmt = create_stmt(token);
	struct statement **stmtp;

	stmt->sym = current_entry->sym;
	stmt->menu = current_entry;
	stmt->prompt = prompt;
	stmt->def = def;
	E_EXPR(stmt->visible) = dep;

	if (prompt)
		current_entry->prompt = stmt;

	/* append statement to the stmt list of symbol */
	if (stmt->sym) {
		for (stmtp = &stmt->sym->stmt; *stmtp; stmtp = &(*stmtp)->next)
			;
		*stmtp = stmt;
	}

	if (!first_stmt)
		first_stmt = stmt;

	return stmt;
}

void menu_finalize(struct menu *parent)
{
	struct menu *menu, *last_menu;
	struct symbol *sym;
	struct statement *stmt;
	struct expr *dep, *dep2;

	//if (parent->sym)
	//	printf("finalize symbol %s\n", parent->sym->name);
	//else
	//	printf("finalize stmt %s\n", parent->prompt->prompt);
	if (!parent->list) {
		sym = parent->sym;
		if (sym) {
			if (sym->type == SYMBOL_UNKNOWN)
				fprintf(stderr, "symbol '%s' has no type\n", sym->name);
			for (stmt = sym->stmt; stmt; stmt = stmt->next) {
				dep = E_EXPR(stmt->visible);
				dep = expr_eliminate_dups(expr_transform(dep));
				if (sym->type != SYMBOL_TRISTATE)
					dep = expr_trans_bool(dep);
				E_EXPR(stmt->visible) = dep;
			}
		}
		if (sym && parent->prompt) {
			last_menu = NULL;
			for (menu = parent->next; menu; menu = menu->next) {
				dep = menu->prompt ? E_EXPR(menu->prompt->visible) : menu->dep;
				//if (!menu->prompt)
				//	break;
				if (!expr_contains_symbol(dep, sym))
					break;
				if (expr_depends_symbol(dep, sym))
					goto next;
				dep = expr_trans_compare(dep, T_UNEQUAL, &symbol_no);
				dep = expr_eliminate_dups(expr_transform(dep));
				dep2 = expr_trans_compare(E_EXPR(parent->prompt->visible), T_UNEQUAL, &symbol_no);
				dep2 = expr_eliminate_dups(expr_transform(dep2));
				expr_eliminate_eq(&dep, &dep2);
				if (!expr_is_yes(dep2))
					break;
			next:
				menu->parent = parent;
				menu_finalize(menu);
				last_menu = menu;
			}
			if (last_menu) {
				parent->list = parent->next;
				parent->next = menu;
				last_menu->next = NULL;
			}
		}
		return;
	}
	sym = parent->sym;
	dep = NULL;
	for (menu = parent->list; menu; menu = menu->next) {
		menu_finalize(menu);
		if (sym && sym_is_choice(sym) && menu->sym) {
			current_entry = parent;
			menu_set_type(menu->sym->type);
			menu->sym->flags |= SYMBOL_CHOICEVAL;
			current_entry = menu;
			menu_set_type(sym->type);
			menu_add_stmt('^', NULL, parent->sym, NULL);
			stmt = sym_get_choice_stmt(parent->sym);
			//dep = expr_alloc_one('^', dep);
			//dep->right.sym = menu->sym;
			stmt->dep = expr_alloc_one('^', stmt->dep);
			stmt->dep->right.sym = menu->sym;
		}
		if (menu->list && (!menu->prompt || !menu->prompt->prompt)) {
			for (last_menu = menu->list; last_menu->next;
			     last_menu = last_menu->next)
				;
			last_menu->next = menu->next;
			menu->next = menu->list;
			menu->list = NULL;
		}
	}
}

bool menu_is_visible(struct menu *menu)
{
	tristate visible;

	if (!menu->prompt)
		return false;
	if (menu->sym) {
		sym_calc_value(menu->sym);
		visible = E_TRI(menu->prompt->visible);
	} else
		visible = E_CALC(menu->prompt->visible);
	return visible != no;
}

char *menu_get_prompt(struct menu *menu)
{
	if (menu->prompt)
		return menu->prompt->prompt;
	else if (menu->sym)
		return menu->sym->name;
	return NULL;
}

struct file *file_lookup(const char *name)
{
	struct file *file;

	for (file = file_list; file; file = file->next) {
		if (!strcmp(name, file->name))
			return file;
	}

	file = malloc(sizeof(*file));
	memset(file, 0, sizeof(*file));
	file->name = strdup(name);
	file->next = file_list;
	file_list = file;
	return file;
}

int file_write_dep(const char *name)
{
	struct file *file;
	FILE *out;

	if (!name)
		name = "..config.cmd";
	out = fopen("..config.tmp", "w");
	if (!out)
		return 1;
	fprintf(out, "deps_config := \\\n");
	for (file = file_list; file; file = file->next) {
		if (file->next)
			fprintf(out, "\t%s \\\n", file->name);
		else
			fprintf(out, "\t%s\n", file->name);
	}
	fprintf(out, "\n.config include/linux/autoconf.h: $(deps_config)\n\n$(deps_config):\n");
	fclose(out);
	rename("..config.tmp", name);
	return 0;
}

