/*
 * Copyright (C) 2002 Roman Zippel <zippel@linux-m68k.org>
 * Released under the terms of the GNU GPL v2.0.
 */

#include <qapplication.h>
#include <qvbox.h>
#include <qsplitter.h>
#include <qlistview.h>
#include <qtextview.h>
#include <qlineedit.h>
#include <qmenubar.h>
#include <qmessagebox.h>
#include <stdlib.h>
#include "zconf.h"
#include "zconf.tab.h"
#include "qconf.h"
#include "version.h"

#include "qconf.moc"

static const char *xpm_symbol_no[] = {
"12 12 2 1",
"  c white",
". c black",
"            ",
" .......... ",
" .        . ",
" .        . ",
" .        . ",
" .        . ",
" .        . ",
" .        . ",
" .        . ",
" .        . ",
" .......... ",
"            "};

static const char *xpm_symbol_mod[] = {
"12 12 2 1",
"  c white",
". c black",
"            ",
" .......... ",
" .        . ",
" .        . ",
" .   ..   . ",
" .  ....  . ",
" .  ....  . ",
" .   ..   . ",
" .        . ",
" .        . ",
" .......... ",
"            "};

static const char *xpm_symbol_yes[] = {
"12 12 2 1",
"  c white",
". c black",
"            ",
" .......... ",
" .        . ",
" .        . ",
" .      . . ",
" .     .. . ",
" . .  ..  . ",
" . ....   . ",
" .  ..    . ",
" .        . ",
" .......... ",
"            "};

static const char *xpm_choice_no[] = {
"12 12 2 1",
"  c white",
". c black",
"            ",
"    ....    ",
"  ..    ..  ",
"  .      .  ",
" .        . ",
" .        . ",
" .        . ",
" .        . ",
"  .      .  ",
"  ..    ..  ",
"    ....    ",
"            "};

static const char *xpm_choice_yes[] = {
"12 12 2 1",
"  c white",
". c black",
"            ",
"    ....    ",
"  ..    ..  ",
"  .      .  ",
" .   ..   . ",
" .  ....  . ",
" .  ....  . ",
" .   ..   . ",
"  .      .  ",
"  ..    ..  ",
"    ....    ",
"            "};

static const char *xpm_menu[] = {
"12 12 2 1",
"  c white",
". c black",
"            ",
" .......... ",
" .        . ",
" . ..     . ",
" . ....   . ",
" . ...... . ",
" . ...... . ",
" . ....   . ",
" . ..     . ",
" .        . ",
" .......... ",
"            "};


/*
 * update all the children of a menu entry
 *   removes/adds the entries from the parent widget as necessary
 *
 * parent: either the menu list widget or a menu entry widget
 * menu: entry to be updated
 */
template <class P>
static void updateMenuList(P* parent, struct menu* menu, bool showAll, bool showFull)
{
	struct menu* child;
	ConfigItem* last;
	bool visible;

	last = 0;
	for (child = menu->list; child; child = child->next) {
		visible = menu_is_visible(child);

		if ((showAll || visible) && showFull) {
			if (!child->data)
				child->data = new ConfigItem(parent, last, child);
			last = (ConfigItem*)child->data;
			last->setEnabled(visible);
			last->updateMenu();
		} else {
			if (child->data) {
				struct menu* m;

				delete (ConfigItem*)child->data;
				// find the previous list entry
				last = 0;
				for (m = menu->list; m != child; m = m->next)
					if (m->data)
						last = (ConfigItem*)m->data;
			}
		}
	}
}

#if QT_VERSION >= 300
/*
 * set the new data
 * TODO check the value
 */
void ConfigItem::okRename(int col)
{
	Parent::okRename(col);
	sym_set_string_value(menu->sym, text(listView()->dataColIdx).latin1());
}
#endif

/*
 * update the displayed of a menu entry
 */
void ConfigItem::updateMenu(void)
{
	ConfigList* list = listView();
	struct symbol* sym;
	int type;
	tristate expr;
	bool showFull = true;

	if (menu->prompt && menu->prompt->token == T_MENU)
		showFull = list->showFull();

	sym = menu->sym;
	if (!sym)
		goto done;

	if (!(sym->flags & SYMBOL_CHANGED) && !list->updateAll)
		goto done;

	sym->flags &= ~SYMBOL_CHANGED;

	setText(list->nameColIdx, menu->sym->name);
	//setText(list->promptColIdx, menu_get_prompt(menu));

	type = sym_get_type(sym);
	switch (type) {
	case SYMBOL_BOOLEAN:
	case SYMBOL_TRISTATE:
		char ch;

		setText(list->promptColIdx, QString().sprintf("%s%s", menu_get_prompt(menu),
				sym_has_value(sym) ? "" : " (NEW)"));
		if (!sym_is_changable(sym) && !list->showAll()) {
			setText(list->noColIdx, 0);
			setText(list->modColIdx, 0);
			setText(list->yesColIdx, 0);
			break;
		}
		expr = sym_get_tristate_value(sym);
		switch (expr) {
		case yes:
			if (sym_is_choice_value(sym) && type == SYMBOL_BOOLEAN)
				setPixmap(list->promptColIdx, list->choiceYesPix);
			else
				setPixmap(list->promptColIdx, list->symbolYesPix);
			setText(list->yesColIdx, "Y");
			ch = 'Y';
			break;
		case mod:
			setPixmap(list->promptColIdx, list->symbolModPix);
			setText(list->modColIdx, "M");
			ch = 'M';
			break;
		default:
			if (sym_is_choice_value(sym) && type == SYMBOL_BOOLEAN)
				setPixmap(list->promptColIdx, list->choiceNoPix);
			else
				setPixmap(list->promptColIdx, list->symbolNoPix);
			setText(list->noColIdx, "N");
			ch = 'N';
			break;
		}
		if (expr != no)
			setText(list->noColIdx, sym_tristate_within_range(sym, no) ? "_" : 0);
		if (expr != mod)
			setText(list->modColIdx, sym_tristate_within_range(sym, mod) ? "_" : 0);
		if (expr != yes)
			setText(list->yesColIdx, sym_tristate_within_range(sym, yes) ? "_" : 0);

#if 0
		if (type == SYMBOL_BOOLEAN)
			setText(list->promptColIdx, QString().sprintf("[%c] %s", ch, menu_get_prompt(menu)));
		else
			setText(list->promptColIdx, QString().sprintf("<%c> %s", ch, menu_get_prompt(menu)));
		setText(list->promptColIdx, menu_get_prompt(menu));
#endif
		setText(list->dataColIdx, QChar(ch));
		break;
	case SYMBOL_INT:
	case SYMBOL_HEX:
	case SYMBOL_STRING:
		const char* data;

#if 0
		if (S_VAL(sym->def))
			data = (char*)S_VAL(sym->def);
		else if (menu->prompt && menu->prompt->def) {
			sym_calc_value(menu->prompt->def);
			data = sym_get_string_value(menu->prompt->def);
		} else
#endif
			data = sym_get_string_value(sym);
#if QT_VERSION >= 300
		setRenameEnabled(list->dataColIdx, TRUE);
#endif
		setText(list->dataColIdx, data);
		if (type == SYMBOL_STRING)
			setText(list->promptColIdx, QString().sprintf("%s: %s", menu_get_prompt(menu), data));
		else
			setText(list->promptColIdx, QString().sprintf("(%s) %s", data, menu_get_prompt(menu)));
		break;
	}
done:
	updateMenuList(this, menu, list->showAll(), showFull);
}

/*
 * construct a menu entry
 */
void ConfigItem::init(void)
{
	ConfigList* list = listView();
#if QT_VERSION < 300
	setEnabled(TRUE);
#endif
	menu->data = this;
	if (menu->sym) {
		menu->sym->flags |= SYMBOL_CHANGED;
	} else {
		setText(list->promptColIdx, menu_get_prompt(menu));
		if (menu->prompt && menu->prompt->token == T_MENU && !list->showFull())
			setPixmap(list->promptColIdx, list->menuPix);
	}

	if (!list->showFull())
		setOpen(TRUE);
}

/*
 * destruct a menu entry
 */
ConfigItem::~ConfigItem(void)
{
	menu->data = 0;
}

void ConfigLineEdit::show(ConfigItem* i)
{
	item = i;
	if (sym_get_string_value(item->menu->sym))
		setText(sym_get_string_value(item->menu->sym));
	else
		setText(0);
	Parent::show();
	setFocus();
}

void ConfigLineEdit::keyPressEvent(QKeyEvent* e)
{
	switch (e->key()) {
	case Key_Escape:
		break;
	case Key_Return:
	case Key_Enter:
		sym_set_string_value(item->menu->sym, text().latin1());
		emit lineChanged(item);
		break;
	default:
		Parent::keyPressEvent(e);
		return;
	}
	e->accept();
	hide();
}

ConfigList::ConfigList(QWidget* p)
	: Parent(p),
	  _showAll(false), _showName(true), _showRange(true), _showData(true), _showFull(true),
	  rootEntry(&rootmenu),
	  promptColIdx(-1), nameColIdx(-1), noColIdx(-1),
	  modColIdx(-1), yesColIdx(-1), dataColIdx(-1),
	  updateAll(false),
	  symbolYesPix(xpm_symbol_yes), symbolModPix(xpm_symbol_mod), symbolNoPix(xpm_symbol_no),
	  choiceYesPix(xpm_choice_yes), choiceNoPix(xpm_choice_no), menuPix(xpm_menu)
{
	setSorting(-1);
	setRootIsDecorated(TRUE);
	promptColIdx = addColumn("prompt");

	connect(this, SIGNAL(clicked(QListViewItem*, const QPoint&, int)),
		SLOT(setValue(QListViewItem*, const QPoint&, int)));

	reinit();
}

void ConfigList::reinit(void)
{
	if (dataColIdx >= 0) {
		removeColumn(dataColIdx);
		dataColIdx = -1;
	}
	if (yesColIdx >= 0) {
		removeColumn(yesColIdx);
		yesColIdx = -1;
	}
	if (modColIdx >= 0) {
		removeColumn(modColIdx);
		modColIdx = -1;
	}
	if (noColIdx >= 0) {
		removeColumn(noColIdx);
		noColIdx = -1;
	}
	if (nameColIdx >= 0) {
		removeColumn(nameColIdx);
		nameColIdx = -1;
	}

	if (showName())
		nameColIdx = addColumn("name");
	if (showRange()) {
		noColIdx = addColumn("n");
		modColIdx = addColumn("m");
		yesColIdx = addColumn("y");
	}
	if (showData())
		dataColIdx = addColumn("data");

	updateListAll();
}

void ConfigList::updateList(ConfigItem* item)
{
	(void)item;	// unused so far
	updateMenuList(this, rootEntry, showAll(), true);
}

void ConfigList::setValue(QListViewItem* i, const QPoint& pnt, int col)
{
	ConfigItem* item;
	struct symbol* sym;
	int type;
	tristate oldval, newval;

	if (!i)
		return;
	item = (ConfigItem*)i;
#if QT_VERSION < 300
	if (!item->enabled)
		return;
#endif
	sym = item->menu->sym;
	if (!sym)
		return;

	type = sym_get_type(sym);
	switch (type) {
	case SYMBOL_BOOLEAN:
	case SYMBOL_TRISTATE:
		oldval = sym_get_tristate_value(sym);
		if (col == noColIdx)
			newval = no;
		else if (col == modColIdx)
			newval = mod;
		else if (col == yesColIdx)
			newval = yes;
		else
			return;

		if (!sym_set_tristate_value(sym, newval))
			return;
		if (oldval == no && item->menu->list)
			item->setOpen(TRUE);
		break;
	case SYMBOL_INT:
	case SYMBOL_HEX:
	case SYMBOL_STRING:
		if (col == dataColIdx) {
#if QT_VERSION >= 300
			item->startRename(dataColIdx);
#else
			lineEdit->show(item);
#endif
		}
		break;
	}

	updateList(item);
}

void ConfigList::changeValue(QListViewItem* i)
{
	ConfigItem* item;
	struct symbol* sym;
	struct menu* menu;
	int type, oldexpr, newexpr;

	if (!i)
		return;

	item = (ConfigItem*)i;
#if QT_VERSION < 300
	if (!item->enabled)
		return;
#endif

	menu = item->menu;
	sym = menu->sym;
	if (!sym) {
		if (item->menu->list)
			item->setOpen(!item->isOpen());
		return;
	}

	type = sym_get_type(sym);
	switch (type) {
	case SYMBOL_BOOLEAN:
	case SYMBOL_TRISTATE:
		oldexpr = sym_get_tristate_value(sym);
		newexpr = sym_toggle_tristate_value(sym);
		if (item->menu->list) {
			if (oldexpr == newexpr)
				item->setOpen(!item->isOpen());
			else if (oldexpr == no)
				item->setOpen(TRUE);
		}
		break;
	case SYMBOL_INT:
	case SYMBOL_HEX:
	case SYMBOL_STRING:
#if QT_VERSION >= 300
		if (dataColIdx >= 0)
			item->startRename(dataColIdx);
		else
#endif
			lineEdit->show(item);
		break;
	}

	updateList(item);
}

void ConfigList::keyPressEvent(QKeyEvent* ev)
{
	QListViewItem* i = currentItem();
	ConfigItem* item;

	if (ev->key() == Key_Escape && !showFull() && rootEntry != &rootmenu) {
		struct menu *oldroot, *newroot;
		oldroot = newroot = rootEntry;
		do {
			newroot = newroot->parent;
			if (newroot->prompt && newroot->prompt->token == T_MENU)
				break;
		} while (newroot != &rootmenu);
		updateMenuList(this, rootEntry, false, false);
		rootEntry = newroot;
		updateListAll();
		setCurrentItem((ConfigItem*)oldroot->data);
		ev->accept();
		return;
	}

	if (!i) {
		Parent::keyPressEvent(ev);
		return;
	}
	item = (ConfigItem*)i;

	switch (ev->key()) {
	case Key_Return:
	case Key_Enter:
		if (item->menu->prompt && item->menu->prompt->token == T_MENU && !showFull()) {
			struct menu *menu = item->menu;
			updateMenuList(this, rootEntry, false, false);
			rootEntry = menu;
			updateListAll();
			break;
		}
	case Key_Space:
		changeValue(i);
		break;
	case Key_N:
		setValue(i, QPoint(), noColIdx);
		break;
	case Key_M:
		setValue(i, QPoint(), modColIdx);
		break;
	case Key_Y:
		setValue(i, QPoint(), yesColIdx);
		break;
	default:
		Parent::keyPressEvent(ev);
		return;
	}
	ev->accept();
}

void ConfigList::setShowAll(bool showAll)
{
	if (showAll == _showAll)
		return;
	_showAll = showAll;
	updateListAll();
	emit showAllChanged(showAll);
}

void ConfigList::setShowName(bool showName)
{
	if (showName == _showName)
		return;
	_showName = showName;
	reinit();
	emit showNameChanged(showName);
}

void ConfigList::setShowRange(bool showRange)
{
	if (showRange == _showRange)
		return;
	_showRange = showRange;
	reinit();
	emit showRangeChanged(showRange);
}

void ConfigList::setShowData(bool showData)
{
	if (showData == _showData)
		return;
	_showData = showData;
	reinit();
	emit showDataChanged(showData);
}

void ConfigList::setShowFull(bool showFull)
{
	if (showFull == _showFull)
		return;
	_showFull = showFull;
	updateMenuList(this, rootEntry, false, false);
	rootEntry = &rootmenu;
	updateListAll();
	emit showFullChanged(showFull);
}

/*
 * Construct the complete config widget
 */
ConfigView::ConfigView(void)
{
	QMenuBar* menu;
	QSplitter* splitter;
	QPopupMenu* config;

	menu = new QMenuBar(this);
	splitter = new QSplitter(this);
	splitter->setOrientation(QSplitter::Vertical);

	// create config menu
	config = new QPopupMenu(this);
	config->setFocusPolicy(QWidget::ClickFocus);
	config->insertItem("&quit", this, SLOT(close()), CTRL+Key_Q);
	menu->insertItem("&config", config);

	// create options menu
	optionMenu = new QPopupMenu(this);
	//optionMenu->setFocusPolicy(QWidget::ClickFocus);
	showFullId = optionMenu->insertItem("full tree", this, SLOT(toggleShowFull()));
	showAllId = optionMenu->insertItem("show all", this, SLOT(toggleShowAll()));
	showNameId = optionMenu->insertItem("show name", this, SLOT(toggleShowName()));
	showRangeId = optionMenu->insertItem("show range", this, SLOT(toggleShowRange()));
	showDataId = optionMenu->insertItem("show data", this, SLOT(toggleShowData()));
	menu->insertItem("&option", optionMenu);


	// create config tree
	QVBox* box = new QVBox(splitter);
	configList = new ConfigList(box);
	configList->lineEdit = new ConfigLineEdit(box);
	configList->lineEdit->hide();
	configList->connect(configList->lineEdit, SIGNAL(lineChanged(ConfigItem*)),
			    SLOT(updateList(ConfigItem*)));

	helpText = new QTextView(splitter);

	setTabOrder(configList, helpText);
	configList->setFocus();

	optionMenu->setItemChecked(showAllId, configList->showAll());
	connect(configList, SIGNAL(showAllChanged(bool)),
		SLOT(updateShowAll(bool)));
	optionMenu->setItemChecked(showNameId, configList->showName());
	connect(configList, SIGNAL(showNameChanged(bool)),
		SLOT(updateShowName(bool)));
	optionMenu->setItemChecked(showRangeId, configList->showRange());
	connect(configList, SIGNAL(showRangeChanged(bool)),
		SLOT(updateShowRange(bool)));
	optionMenu->setItemChecked(showDataId, configList->showData());
	connect(configList, SIGNAL(showDataChanged(bool)),
		SLOT(updateShowData(bool)));
	optionMenu->setItemChecked(showFullId, configList->showFull());
	connect(configList, SIGNAL(showFullChanged(bool)),
		SLOT(updateShowFull(bool)));
	connect(configList, SIGNAL(selectionChanged(QListViewItem*)),
		SLOT(setHelp(QListViewItem*)));
}

/*
 * display a new help entry as soon as a new menu entry is selected
 */
void ConfigView::setHelp(QListViewItem* item)
{
	struct symbol* sym;
	struct menu* menu;

	configList->lineEdit->hide();
	if (item) {
		menu = ((ConfigItem*)item)->menu;

		sym = menu->sym;
		if (sym) {
			helpText->setText(sym->help);
			return;
		}
	}
	helpText->setText(NULL);
}

/*
 * show/hide menu entries without prompt
 */
void ConfigView::toggleShowAll(void)
{
	configList->setShowAll(!optionMenu->isItemChecked(showAllId));
}

void ConfigView::updateShowAll(bool showAll)
{
	optionMenu->setItemChecked(showAllId, showAll);
}

void ConfigView::toggleShowName(void)
{
	configList->setShowName(!optionMenu->isItemChecked(showNameId));
}

void ConfigView::updateShowName(bool showName)
{
	optionMenu->setItemChecked(showNameId, showName);
}

void ConfigView::toggleShowRange(void)
{
	configList->setShowRange(!optionMenu->isItemChecked(showRangeId));
}

void ConfigView::updateShowRange(bool showRange)
{
	optionMenu->setItemChecked(showRangeId, showRange);
}

void ConfigView::toggleShowData(void)
{
	configList->setShowData(!optionMenu->isItemChecked(showDataId));
}

void ConfigView::updateShowData(bool showData)
{
	optionMenu->setItemChecked(showDataId, showData);
}

void ConfigView::toggleShowFull(void)
{
	configList->setShowFull(!optionMenu->isItemChecked(showFullId));
}

void ConfigView::updateShowFull(bool showFull)
{
	optionMenu->setItemChecked(showFullId, showFull);
}


/*
 * ask for saving configuration before quitting
 * TODO ask only when something changed
 */
void ConfigView::closeEvent(QCloseEvent* e)
{
	QMessageBox mb("qconf", "Save configuration?", QMessageBox::Warning,
			QMessageBox::Yes | QMessageBox::Default, QMessageBox::No, QMessageBox::Cancel | QMessageBox::Escape);
	switch (mb.exec()) {
	case QMessageBox::Yes:
		conf_write(NULL);
	case QMessageBox::No:
		e->accept();
		break;
	case QMessageBox::Cancel:
		e->ignore();
		break;
	}
}

int main(int ac, char** av)
{
	QApplication a(ac, av);
	ConfigView* v;
	const char *name;

	if (ac > 1 && av[1][0] == '-') {
		switch (av[1][1]) {
		case 'a':
			//showAll = 1;
			break;
		case 'h':
		case '?':
			printf("%s -a\n", av[0]);
			exit(0);
		}
		name = av[2];
	} else
		name = av[1];
	conf_parse(name);
	conf_read(NULL);
	//zconfdump(stdout);
	v = new ConfigView();

	//zconfdump(stdout);
	v->show();
	a.connect(&a, SIGNAL(lastWindowClosed()), &a, SLOT(quit()));
	a.exec();
	return 0;
}
